from jsonpath_ng import parse
from .paginate import ddb_paginate

#####################################

class list_tables(ddb_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('TableNames[*]')


from jsonpath_ng import parse
from .paginate import ec_paginate

#####################################

class describe_cache_clusters(ec_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('CacheClusters[*]')

#####################################

class describe_reserved_cache_nodes(ec_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('ReservedCacheNodes[*]')


from jsonpath_ng import parse
from .paginate import autoscaling_paginate

#####################################

class describe_auto_scaling_groups(autoscaling_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('AutoScalingGroups[*]')

#####################################

class describe_policies(autoscaling_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('ScalingPolicies[*]')

#####################################



from jsonpath_ng import parse
from .paginate import ec2_paginate

#####################################

class describe_regions(ec2_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('Regions[*]')

#####################################

class describe_instances(ec2_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('Reservations[*].Instances[*]')

#####################################

class describe_vpc_endpoints(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('VpcEndpoints[*]')

#####################################

class describe_nat_gateways(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('NatGateways[*]')

#####################################

class describe_subnets(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('Subnets[*]')

#####################################

class describe_route_tables(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('RouteTables[*]')

#####################################

class describe_prefix_lists(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('PrefixLists[*]')

#####################################

class describe_vpcs(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('Vpcs[*]')

#####################################

class describe_instance_types(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('InstanceTypes[*]')

#####################################

class describe_launch_templates(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('LaunchTemplates[*]')

#####################################

class describe_launch_template_versions(ec2_paginate):
    def name(self): return __class__.__name__
    def jsonpath_expression(self): return parse('LaunchTemplateVersions[*]')

#####################################

import boto3
import datetime as dt  
from itertools import groupby
from datetime import datetime, timedelta

def get_cpu_util_metrics(**kargs):
    
    client = kargs.get('session').client('cloudwatch')
    iid = kargs.get('instance_id')
    days = kargs.get('days',31)
    period = kargs.get('period',300)
    dimensions = [{"Name": "InstanceId", "Value": iid}]

    dimensions = [{"Name": "InstanceId", "Value": instance_id}]
    queries = [
        {
            "Id": "cpu_util", 
            "MetricStat": {
                "Metric": {
                    "Namespace": "AWS/EC2", 
                    "MetricName": "CPUUtilization", 
                    "Dimensions": dimensions
                },                
                "Period": period,                
                "Stat": "Maximum"
            },            
            "Label": "cpu_util_max",        
        },
        {
            "Id": "cpu_util", 
            "MetricStat": {
                "Metric": {
                    "Namespace": "AWS/EC2", 
                    "MetricName": "CPUUtilization", 
                    "Dimensions": dimensions
                },                
                "Period": period,                
                "Stat": "SampleCount"
            },            
            "Label": "cpu_util_count",        
        },

    ]
    
    results = client.get_metric_data(
        MetricDataQueries=queries,
        StartTime = dt.datetime.today() - timedelta(days=days + 1),
        EndTime=dt.datetime.today() - timedelta(days=1),
    )["MetricDataResults"]
    
    return {k: list(group) for k, group in groupby(results, key=lambda r: r["Label"])}

#####################################




import os
import json
import asyncio
import pandas as pd
from typing import Any
from itertools import groupby
from functools import wraps, partial
from concurrent.futures import ThreadPoolExecutor

MAX_WORKERS = int(os.environ.get('MAX_WORKERS', '32'))
worker_pool = ThreadPoolExecutor(max_workers=MAX_WORKERS)

def worker(func: Any) -> Any:
    """
    Runs blocking code asynchronously in worker thread from worker pool
    """
    @wraps(func)
    async def run(*args, **kwargs):
        pfunc = partial(func, *args, **kwargs)
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(worker_pool, pfunc)
    return run


@worker
def gather_data(account, region, instances):
    pass

def grouped_instances(records, account="account_id", region="region"):
    grouped_instances = groupby(records, lambda r: (r[account], r[region]))
    return {key: list(group) for key, group in grouped_instances}

def create_tasks(grouped_instances, func=gather_data):
    grouped_instances_dict = {key: list(group) for key, group in grouped_instances.items()}
    return [func(account, region, instances) for (account, region), instances in grouped_instances_dict.items()]    

def updated_dataframe(records, groupd_instances):
    return pd.DataFrame([record for records in grouped_instances.values() for record in records])


def call_endpoint(accounts, sessions, **kargs):
    data = {'accounts':accounts, 'sessions':sessions}
    return [run_func(a, data, **kargs) for a in accounts]

def run_func(a, data, **kargs):
    aid = a.get('account_id')
    region = a.get('region')
    s = data['sessions'].get(aid,{}).get(region,None)
    if s: 
        kargs.update({'row':a})
        a.update({kargs.get('key'): kargs.get('func')(s, **kargs)})
        return a
    return None     



import boto3
import pandas as pd
from io import StringIO as sio

#####################################

def get_assumed_object(s, **kargs):
    s3 = s.client('s3', kargs.get('region','us-east-1'))
    obj = s3.get_object(Bucket=kargs.get('bucket'), Key=kargs.get('key'))
    return obj['Body'].read().decode('utf-8') 

def get_object(**kargs):
    s3 = boto3.resource('s3')
    obj = s3.Object(kargs.get('bucket'), kargs.get('key'))
    return obj.get()['Body'].read().decode('utf-8') 

def get_csv_data_from_s3(**kargs):
    s = kargs.get('session')
    b = kargs.get('bucket')
    k = kargs.get('key')
    data = get_assumed_object(s, bucket=b, key=k)
    return pd.read_csv(sio(data), sep=","); 

#####################################

def put_object(s, **kargs):
    d = kargs.get('data')
    k = kargs.get('key')
    b = kargs.get('bucket')
    r = kargs.get('region', 'us-east-1')
    s.client('s3', r).put_object(Body=d, Bucket=b, Key=k)
    print(f'https://{b}.s3.amazonaws.com/{k}')

#####################################



import time
import awswrangler as wr
from functools import partial
import pandas as pd

#####################

def query(session, s, **kargs):
    generated_query = generate_query(s, **kargs) 
    return partial(wr.athena.read_sql_query,
                   database=kargs.get('db'),
                   boto3_session=session)(generated_query)

#####################

def generate_query(s, **kargs):
    for key,val in kargs.items():
        s = s.replace('{' + key + '}', str(val)) if f'{key}' in s else s
    return s

#####################

def execute_query(s,query, db, wg):
    try:
        athena = s.client('athena')
        query_execution_id = athena.start_query_execution(
            QueryString=query, QueryExecutionContext={"Database": db}, WorkGroup=wg
        )["QueryExecutionId"]

        # Poll for result
        result = athena.get_query_execution(QueryExecutionId=query_execution_id)["QueryExecution"]
        while result["Status"]["State"] in ["QUEUED", "RUNNING"]:
            result = athena.get_query_execution(QueryExecutionId=query_execution_id)["QueryExecution"]
            time.sleep(1)
                    
        # Copy result to dataframe
        if result["Status"]["State"] == "SUCCEEDED":
            bucket, key = result["ResultConfiguration"]["OutputLocation"].split("s3://")[-1].split("/", 1)
            body = s.client('s3').get_object(Bucket=bucket, Key=key)["Body"]
            return pd.read_csv(body).reset_index(drop=True)
        else:
            return pd.DataFrame()
    except Exception as e:
        print(e) 
        return pd.DataFrame()


import boto3

#####################

def session_from_credentials(creds, region):
    try:
        return boto3.session.Session(
            region_name=region,
            aws_access_key_id=creds["AccessKeyId"],
            aws_secret_access_key=creds["SecretAccessKey"],
            aws_session_token=creds["SessionToken"]
        )
    except:
        return None

#####################

def assume_role(**params):
    region = params.get('region', 'us-east-1')
    access_key_id = params.get('access_key_id')
    secret_access_key = params.get('secret_access_key') 
 
    account_id = params.get('accountId')
    role_name = params.get('roleName')
    session_name = params.get('sessionName','tmp')
    role_arn = f'arn:aws:iam::{account_id}:role/{role_name}'

    try:
        if access_key_id:
            sts = boto3.client('sts', 
                               region_name=region,
                               aws_access_key_id=access_key_id,                
                               aws_secret_access_key=secret_access_key)
        else:
            sts = boto3.client('sts')
        assumed_role = sts.assume_role(RoleArn=role_arn, 
                                       RoleSessionName=session_name)
        credentials = assumed_role.get('Credentials')
    except Exception as e:
        print(account_id, 'Assume error')
        credentials = None

    return session_from_credentials(credentials, region)

############################################

def get_account_regions(accounts):
    account_regions = list(set([(a['account_id'], a['region']) for a in accounts]))
    return [{'account_id':a[0], 'region':a[1]} for a in account_regions]
    
#####################

def get_session(a, **args):
    reg = a.get('region')
    aid = a.get('account_id')
    sessions = args.get('sessions')
    if not sessions: 
        for ar in args.get('account_regions'): sessions.update({ar['account_id']:{}})
    s = sessions.get(aid,{}).get(reg,None)
    if not s: 
        # Create a new session object and cache it in the sessions dictionary
        s = assume_role(access_key_id = args.get('key'), 
                        secret_access_key = args.get('secret'),
                        roleName = args.get('role'),
                        accountId = aid, region=reg)
        sessions[aid][reg] = s
    return s
        
#####################

def get_sessions(accounts, **kargs):
    account_regions = get_account_regions(accounts)
    kargs.update({'account_regions': account_regions, 'sessions': {}})
    for a in accounts:
        account_id = a['account_id']
        region = a['region']
        # Check if the account/region pair has already been processed
        if (account_id, region) in kargs['sessions']:
            continue
        # Call get_session() and add the resulting session to the sessions dictionary
        session = get_session(a, **kargs)
        kargs['sessions'][(account_id, region)] = session
    return kargs['sessions']



    #########################################

    import boto3
from saas_co import session as cos

def get_main_sts_client(secret_id, access_key_id, secret_access_key, region = 'us-east-1'):
    sm_client = boto3.client('secretsmanager')
    secretString = eval(sm_client.get_secret_value(SecretId=secret_id)['SecretString'])
    access_key_id = secretString[access_key_id]; 
    secret_access_key = secretString[secret_access_key]
    
    return boto3.client(
        'sts', 
        region_name=region,
        aws_access_key_id=access_key_id,                
        aws_secret_access_key=secret_access_key)

def get_main_session(sts_client, account, role, region, session_name):
    creds = sts_client.assume_role(
        RoleArn=f'arn:aws:iam::{account}:role/{role}',
        RoleSessionName=session_name).get('Credentials')
    return cos.session_from_credentials(creds, region)
    
def get_session_with_external_id(sts_client, account, role, region, session_name, external_id):
    creds = sts_client.assume_role(
        RoleArn=f'arn:aws:iam::{account}:role/{role}',
        RoleSessionName=session_name, ExternalId=external_id).get('Credentials')
    return cos.session_from_credentials(creds, region)
    
def get_main_power_session(secret_id, access_key_id, secret_access_key, account, region, role, session_name):
    sts_client = get_main_sts_client(secret_id, access_key_id, secret_access_key, region)
    return get_main_session(sts_client, account, role, region, session_name)

##########################

def get_sts_client_from_session(s, account, role, region):
    creds = s.get_credentials().get_frozen_credentials()
    return boto3.client('sts',
        aws_access_key_id=creds.access_key,
        aws_secret_access_key=creds.secret_key,
        aws_session_token=creds.token,
        region_name=region)

def get_session_from_session(s, account, role, region):
    sts_client = get_sts_client_from_session(s, account, role, region)
    session_name = s.client('sts').get_caller_identity()['Arn'].split('/')[-1].split(':')[0]    
    return get_main_session(sts_client, account, role, region, session_name)

def get_session_from_session_with_accounts(s, account, role, region, accounts):
    external_id = get_external_id_from_account(s, accounts, account)
    sts_client = get_sts_client_from_session(s, account, role, region)
    session_name = s.client('sts').get_caller_identity()['Arn'].split('/')[-1].split(':')[0]
    return get_session_with_external_id(sts_client, account, role, region, session_name, external_id)

def get_main_finder_session(s, account, region, role):
    try:     return get_session_from_session(s, account, role, region)
    except:  return None
    
def get_customer_session(s, accounts, account, region, role):
    try:    return get_session_from_session_with_accounts(s, account, role, region, accounts)
    except Exception as e:
        print(e)
        return None
    
##########################################################################
##########################################################################

def get_tid_by_account(accounts, account):
    filtered = accounts[accounts['account'].astype(str).apply(lambda x: str(account) in x)]
    return None if filtered.empty else filtered.iloc[0]['tid']

########################

def get_external_id_from_account(main_session, accounts, account):
    tenant_id = get_tid_by_account(accounts, account)
    try:
        response = main_session.client('secretsmanager').get_secret_value(SecretId=tenant_id)
        return response.get('SecretString',response.get('SecretBinary'))
    except Exception as e:
        print(e)
        return None

##########################################################################
##########################################################################







from jsonpath_ng import parse
from .paginate import dms_paginate

#####################################

class describe_replication_instances(dms_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('ReplicationInstances[*]')

#####################################

class describe_replication_tasks(dms_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('ReplicationTasks[*]')

#####################################



from jsonpath_ng import parse
from .paginate import cw_paginate

from datetime import datetime as dt
from datetime import timedelta as td

#####################################

class get_metric_data(cw_paginate):
    def name(self): return __class__.__name__        
    def jsonpath_expression(self): return parse('MetricDataResults[*]')

#####################################

def _get_metric_stat(**kargs):

  stat = kargs.get('Stat')

  try:
      md = kargs.get('session').client('cloudwatch').get_metric_statistics(
                                                        StartTime=dt.now() - td(days=kargs.get('days')), 
                                                        EndTime=dt.now(), 
                                                        Namespace=f"AWS/{kargs.get('Namespace')}",
                                                        MetricName=kargs.get('MetricName'),
                                                        Dimensions=[kargs.get('Dimensions')],
                                                        Period=kargs.get('Period'), 
                                                        Statistics=[stat])
      return md['Datapoints'][0][stat] if len(md['Datapoints']) > 0 else None
  except Exception as e:
        print(e)
        return None

#####################################
# Utility
#####################################

def _get_metric_data(**kargs):

    try:
        return get_metric_data(kargs.get('session'), 
            StartTime=dt.now()-td(days=kargs.get('days')), 
            EndTime=dt.now(), 
            MetricDataQueries=kargs.get('MetricDataQueries'),
            fields=kargs.get('fields',[]));
    except Exception as e:
        print(e)
        return None
    


#####################################

def _get_metric_data_queries(**kargs):
    dims = kargs.get('Dimensions')
    return [{
         "Id": f"{metric}_{dims['Value']}", 
         "MetricStat": {
             "Metric": {
                 "Namespace": f"AWS/{kargs.get('Namespace')}",
                 "MetricName": metric,
                 "Dimensions": [dims]
             },
             "Period": kargs.get('Period'), 
             "Stat": kargs.get('Stat'),
             "Unit": kargs.get("Unit"),
         },
         'Label': metric
      } for metric in kargs.get('MetricNames')]


#####################################

import boto3
import datetime as dt  
from itertools import groupby
from datetime import datetime, timedelta

def get_metrics(**kargs):
    
    client = kargs.get('session').client('cloudwatch')
    iid = kargs.get('instance_id')
    days = kargs.get('days',31)
    period = kargs.get('period',300)
    dimensions = [{"Name": "InstanceId", "Value": iid}]
    queries = [
        {
            "Id": f'id_{i}', 
            "MetricStat": { 
                "Metric": { 
                    "Namespace": nameSpace, 
                    "MetricName": metricName, 
                    "Dimensions": dimensions
                },                
                "Period": period,                
                "Stat": stat,
                "Unit": unit
            },     
            "ReturnData": returnData,
            "Label": label,        
        } for i, (nameSpace, metricName, stat, unit, returnData, label) in enumerate(kargs.get('queries'))
    ]
    
    results = client.get_metric_data(
        MetricDataQueries=queries,
        StartTime = dt.datetime.today() - timedelta(days=days + 1),
        EndTime=dt.datetime.today() - timedelta(days=1),
    )["MetricDataResults"]
    
    return {k: list(group) for k, group in groupby(results, key=lambda r: r["Label"])}      



from .cw import get_metrics

#####################################

def get_ebs_metrics(**kargs):

    kargs['query'] = [
        ("AWS/EC2", "EBSReadBytes",  "Sum", "Bytes", False, "ebs_read_bytes"),
        ("AWS/EC2", "EBSWriteBytes",  "Sum", "Bytes", False, "ebs_write_bytes")
    ]
    return cs.get_metrics(kargs) 

#####################################